import greenfoot.*;
import java.util.Calendar;

/**
 * CLASS:  CalendarActor
 * AUTHOR: danpost (greenfoot.org username)
 * DATE:  November 28, 2018
 * 
 * DESCRIPTION: a monthly calendar that goes back to year 1 and goes forward a lot;
 * it displays currently set day of week and date at top and shows that dates
 * calendar month below;
 * 
 * NOTE: use removeFromWorld method of CalendarActor to remove instance from world
 * to ensure its components are also removed; also, this implementation does not
 * allow CalendarActor object to be moved around in the world, so as a way around
 * this, remove it and add it back in at new location.
 */
public class CalendarActor extends Actor
{
    /** current set date values array { MM, DD, YYYY } initialize to current day */
    private static final int[] FTYPE = { Calendar.MONTH, Calendar.DAY_OF_MONTH, Calendar.YEAR };
    
    /** constant string values */
    private static final String[] DOW = { "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" };
    private static final String[] MOY = { "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December" };
    
    /** helper java.util.Calendar object */
    private Calendar calendar;
    
    /** numbered day actors for calendar cells */
    private DayActor[] das = new DayActor[31];
    
    /** calendar header with "MONTH YYYY" text */
    private Aktor pHeader;
    
    /** actors displaying currently set date values { MM, DD, YYYY } */
    private Actor[] flds = new Actor[3];
    
    /** actor displaying abbreviated day of week of currently set date */
    private Actor dayOfWeek;
    
    /** not currently focused on */
    private int focus = -1;
    
    /** array of String text messages for both states of having focus */
    private String[] msgs = { "Click to activate", "Arrows or 'MDY+-' or click" };
    
    // builds the calendar
    public CalendarActor()
    {
        // the background
        GreenfootImage bg = new GreenfootImage(279, 314);
        bg.setColor(Color.LIGHT_GRAY);
        bg.fill();
        bg.setColor(new Color(224, 224, 224));
        bg.fillRect(0, 71, 279, 30);
        bg.setColor(Color.BLACK);
        bg.drawRect(1, 70, 276, 241);
        bg.drawRect(1, 1, 276, 101);
        bg.setColor(Color.RED);
        bg.drawRect(0, 0, 278, 313);
        // the day of week column headers
        for (int i=0; i<DOW.length; i++)
        {
            GreenfootImage day = new GreenfootImage(DOW[i].substring(0, 3), 21, Color.BLACK, new Color(0, 0, 0, 0));
            bg.drawImage(day, 25+38*i-day.getWidth()/2, 106);
        }
        setImage(bg);
        bg.setColor(Color.LIGHT_GRAY);
        
        // initialize java.util.Calendar object
        calendar = Calendar.getInstance();
    }
    
    /** adds calendar components into world and update to set date */
    protected void addedToWorld(World world)
    {
        // set day of week actor
        dayOfWeek = new Aktor();
        world.addObject(dayOfWeek, getX()-90, getY()+getImage().getHeight()/2-264);
        
        // set day value actors (mm, dd, yyyy)
        for (int i=0; i<3; i++)
        {
            flds[i] = new Aktor();
            world.addObject(flds[i], getX()-30+44*i+14*(i==2?1:0), getY()+getImage().getHeight()/2-264);
            updateField(i); // initializes field value and its display
        }
        
        // initial focus on month
        adjustFocus(0);
        
        // numbered day calendar cell actors
        for (int i=0; i<31; i++)
        {
            das[i] = new DayActor(i+1);
            world.addObject(das[i], 0, 0);
        }
        
        // instruction actor
        pHeader = new Aktor();
        world.addObject(pHeader, getX(), getY()+getImage().getHeight()/2-227);
        
        // update calendar to currently set date
        update();
    }
    
    public void removeFromWorld()
    {
        if (getWorld() == null) return;
        World w = getWorld();
        w.removeObjects(w.getObjects(CalendarActor.DayActor.class));
        w.removeObjects(w.getObjects(CalendarActor.Aktor.class));
        w.removeObject(this);
    }
    
    /** actions changing focus or changing set date */
    public void act()
    {
        // gaining, losing or changing focus by click
        if (Greenfoot.mouseClicked(null))
        {
            // clicks on focusable actor
            Actor actor = Greenfoot.getMouseInfo().getActor();
            if (actor == flds[0]) { adjustFocus(0); return; }
            if (actor == flds[1]) { adjustFocus(1); return; }
            if (actor == flds[2]) { adjustFocus(2); return; }
            // clicks on/off calendar or one of its non-focusable components
            if ((focus >= 0 && 
                (actor == null ||
                    !(actor == this ||
                      actor instanceof CalendarActor.Aktor ||
                      actor instanceof CalendarActor.DayActor)))
             || (focus < 0 &&
                (actor != null &&
                    (actor == this ||
                     actor instanceof CalendarActor.Aktor ||
                     actor instanceof CalendarActor.DayActor))))
                        { adjustFocus(-focus-1); return; }
        }
        // exit if not focus on
        if (focus < 0) return;
        // changing focus by key accelerators
        if (focus != 0 && Greenfoot.isKeyDown("m")) { adjustFocus(0); return; }
        if (focus != 1 && Greenfoot.isKeyDown("d")) { adjustFocus(1); return; }
        if (focus != 2 && Greenfoot.isKeyDown("y")) { adjustFocus(2); return; }
        // changing focus or values by arrow keys 
        String key = Greenfoot.getKey();
        if ("left".equals(key)) { adjustFocus((focus+2)%3); return; }
        if ("right".equals(key)) { adjustFocus((focus+1)%3); return; }
        if ("up".equals(key) || "+".equals(key)) { adjustField(1); return; }
        if ("down".equals(key) || "-".equals(key)) { adjustField(-1); return; }
    }
    
    /** changes currently set date */
    private void adjustField(int dir)
    {
        if (focus == 2 && getYear() == 1 && dir == -1) return;
        calendar.roll(FTYPE[focus], dir);
        for (int i=0; i<3; i++) updateField(i);
        update();
    }
    
    /** removes focus or sets it on a specific elemennt */
    private void adjustFocus(int fnum)
    {
        int hold = focus;
        focus = fnum;
        updateField(hold < 0 ? -hold-1 : hold);
        updateField(focus < 0 ? -focus-1 : focus);
        if (hold < 0 && focus >= 0)
        {
            while (Greenfoot.getKey() != null) ;
            getImage().fillRect(2, 2, 275, 30);
            GreenfootImage text = new GreenfootImage(msgs[1], 24, new Color(0, 128, 0), Color.LIGHT_GRAY);
            getImage().drawImage(text, (getImage().getWidth()-text.getWidth())/2, 4);
        }
        else if (hold >= 0 && focus < 0)
        {
            getImage().fillRect(2, 2, 275, 30);
            GreenfootImage text = new GreenfootImage(msgs[0], 24, Color.RED, Color.LIGHT_GRAY);
            getImage().drawImage(text, (getImage().getWidth()-text.getWidth())/2, 4);
        }
    }
    
    /** updates calendar to currently set date */
    private void update()
    {
        int dom = getDayOfMonth(); // day of month
        calendar.roll(Calendar.DAY_OF_MONTH, 1-dom); // goto first of month
        int idow = getDayOfWeek(); // initial day of week on first of month
        calendar.roll(Calendar.DAY_OF_MONTH, -1); // goto last of month
        int tdom = getDayOfMonth(); // total days in this month
        calendar.roll(Calendar.DAY_OF_MONTH, dom); // restore day
        // position day actors and hide those not included this month
        for (int i=0; i<31; i++)
        {
            das[i].setPosition(getX()-getImage().getWidth()/2+25+38*((idow+i)%7), getY()+getImage().getHeight()/2-201+30*((i+7+idow)/7));
            das[i].update(dom == i+1 ? new Color(176, 255, 176) : Color.WHITE);
            das[i].getImage().setTransparency(i < tdom ? 255 : 0);
        }
        // set header text image
        String title = getMonthName()+"  "+getYear();
        pHeader.setImage(new GreenfootImage(title, 30, Color.BLACK, new Color(0, 0, 0, 0)));
        // set day of week text image
        GreenfootImage img = new GreenfootImage(getDayOfWeekAbbr(), 30, Color.BLACK, new Color(0, 0, 0, 0));
        dayOfWeek.setImage(img);
    }
    
    /** updates a specific displayed value of the set date */
    private void updateField(int fnum)
    {
        GreenfootImage img = new GreenfootImage(40, 30);
        int val = 0;
        switch (fnum)
        {
            case 0: val = getMonth(); break;
            case 1: val = getDayOfMonth(); break;
            case 2: img.scale(68, 30); val = getYear(); break;
        };
        img.setColor(fnum == focus ? new Color(255, 255, 128) : Color.WHITE);
        img.fill();
        img.setColor(Color.BLACK);
        img.drawRect(0, 0, img.getWidth()-1, img.getHeight()-1);
        GreenfootImage text = new GreenfootImage(""+val, 30, Color.BLACK, new Color(0, 0, 0, 0));
        img.drawImage(text, img.getWidth()-6-text.getWidth(), 0);
        flds[fnum].setImage(img);
    }
    
    /** gets the day value of set date */
    public int getDayOfMonth() { return calendar.get(Calendar.DAY_OF_MONTH); }
    
    /** gets the day of week of set date */
    public int getDayOfWeek() { return calendar.get(Calendar.DAY_OF_WEEK)-1; }
    
    /** gets the string text of day of week of set date */
    public String getDayOfWeekName() { return DOW[getDayOfWeek()]; }
    
    /** gets the abbreviated day of week of set date */
    public String getDayOfWeekAbbr() { return getDayOfWeekName().substring(0, 3); }
    
    /** gets the month value of set date */
    public int getMonth() { return calendar.get(Calendar.MONTH)+1; }
    
    /** gets the string text of month of set date */
    public String getMonthName() { return MOY[getMonth()-1]; }
    
    /** gets the abbreviated month of set date */
    public String getMonthAbbr() { return getMonthName().substring(0, 3); }
    
    /** gets the year value of set date */
    public int getYear() { return calendar.get(Calendar.YEAR); }
    
    /** gets the set date as string formatted as "MM?DD?YY" ("?" is divider) */
    public String getMMDDYY(String divider) { return formString(getMonth(), 2, "0")+divider+formString(getDayOfMonth(), 2, "0")+divider+formString(getYear(), 2, ""); }
    
    /** gets the set date as string formatted as "DD?MM?YY" ("?" is divider) */
    public String getDDMMYY(String divider) { return formString(getDayOfMonth(), 2, "0")+divider+formString(getMonth(), 2, "0")+divider+formString(getYear(), 2, ""); }
    
    /** gets millisecond time value at 12:00am of set date */
    public long getDateInMillis() { return calendar.getTimeInMillis()/86400000*86400000; }
    
    /** internal method that returns formed strings (for formatting them) */
    private String formString(int val, int len, String leading)
    {
        String leader = "";
        for (int i=0; i<len; i++) leader += leading;
        String formed = leading+val;
        return formed.substring(formed.length()-len);
    }
    
    /** locks actor in placee (cannot be moved) */
    public void setLocation(int x, int y) {}
    
    /**
     * CLASS: DayActor
     * DESCRIPTION: a day entry for a calendar month
     */
    private class DayActor extends Actor
    {
        int val; // day of month
        
        /** sets image */
        public DayActor(int dom)
        {
            val = dom;
            update(Color.WHITE);
        }
        
        /** updates image (for highlighting purposes) */
        public void update(Color bg)
        {
            GreenfootImage box = new GreenfootImage(37, 30);
            box.setColor(bg);
            box.fill();
            box.setColor(Color.BLACK);
            box.drawRect(0, 0, box.getWidth()-1, box.getHeight()-1);
            GreenfootImage num = new GreenfootImage(""+val, 30, Color.BLACK, new Color(0, 0, 0, 0));
            box.drawImage(num, box.getWidth()-num.getWidth()-6, 1);
            setImage(box);
        }
        
        public void setLocation(int x, int y) {}
        
        protected void setPosition(int x, int y)
        {
            super.setLocation(x, y);
        }
    }
    
    /**
     * CLASS: Aktor
     * DESCRIPTION: a generic actor used for non-moving components
     */
    protected class Aktor extends Actor
    {
        public void setLocation(int x, int y) {}
    }
}